import { CommandHandler, ICommandHandler } from '@nestjs/cqrs';
import { CreateAdCommand } from './create-ad.command';
import { Inject } from '@nestjs/common';
import {
  AD_MESSAGE_PUBLISHER,
  AD_REPOSITORY,
  AD_ROUTE_PROVIDER,
} from '@modules/ad/ad.di-tokens';
import { AdEntity } from '@modules/ad/core/domain/ad.entity';
import { AdRepositoryPort } from '../../ports/ad.repository.port';
import {
  AggregateID,
  ConflictException,
  MessagePublisherPort,
} from '@mobicoop/ddd-library';
import { AdAlreadyExistsException } from '@modules/ad/core/domain/ad.errors';
import { Role } from '@modules/ad/core/domain/ad.types';
import {
  Path,
  PathCreator,
  PathType,
  TypedRoute,
} from '@modules/ad/core/domain/path-creator.service';
import { Waypoint } from '../../types/waypoint.type';
import { Point as PointValueObject } from '@modules/ad/core/domain/value-objects/point.value-object';
import { Point } from '@modules/geography/core/domain/route.types';
import { MatcherAdCreationFailedIntegrationEvent } from '../../events/matcher-ad-creation-failed.integration-event';
import { MATCHER_AD_CREATION_FAILED_ROUTING_KEY } from '@src/app.constants';
import { GeorouterPort } from '../../ports/georouter.port';

@CommandHandler(CreateAdCommand)
export class CreateAdService implements ICommandHandler {
  constructor(
    @Inject(AD_MESSAGE_PUBLISHER)
    private readonly messagePublisher: MessagePublisherPort,
    @Inject(AD_REPOSITORY)
    private readonly repository: AdRepositoryPort,
    @Inject(AD_ROUTE_PROVIDER)
    private readonly routeProvider: GeorouterPort,
  ) {}

  async execute(command: CreateAdCommand): Promise<AggregateID> {
    const roles: Role[] = [];
    if (command.driver) roles.push(Role.DRIVER);
    if (command.passenger) roles.push(Role.PASSENGER);

    const pathCreator: PathCreator = new PathCreator(
      roles,
      command.waypoints.map(
        (waypoint: Waypoint) =>
          new PointValueObject({
            lon: waypoint.lon,
            lat: waypoint.lat,
          }),
      ),
    );

    let typedRoutes: TypedRoute[];
    let driverDistance: number | undefined;
    let driverDuration: number | undefined;
    let passengerDistance: number | undefined;
    let passengerDuration: number | undefined;
    let points: PointValueObject[] | undefined;
    let fwdAzimuth: number | undefined;
    let backAzimuth: number | undefined;

    try {
      try {
        typedRoutes = await Promise.all(
          pathCreator.getBasePaths().map(async (path: Path) => ({
            type: path.type,
            route: await this.routeProvider.getRoute({
              waypoints: path.waypoints,
            }),
          })),
        );
      } catch (e: any) {
        throw new Error('Unable to find a route for given waypoints');
      }

      try {
        typedRoutes.forEach((typedRoute: TypedRoute) => {
          if ([PathType.DRIVER, PathType.GENERIC].includes(typedRoute.type)) {
            driverDistance = typedRoute.route.distance;
            driverDuration = typedRoute.route.duration;
            points = typedRoute.route.points.map(
              (point: Point) =>
                new PointValueObject({
                  lon: point.lon,
                  lat: point.lat,
                }),
            );
            fwdAzimuth = typedRoute.route.fwdAzimuth;
            backAzimuth = typedRoute.route.backAzimuth;
          }
          if (
            [PathType.PASSENGER, PathType.GENERIC].includes(typedRoute.type)
          ) {
            passengerDistance = typedRoute.route.distance;
            passengerDuration = typedRoute.route.duration;
            if (!points)
              points = typedRoute.route.points.map(
                (point: Point) =>
                  new PointValueObject({
                    lon: point.lon,
                    lat: point.lat,
                  }),
              );
            if (!fwdAzimuth) fwdAzimuth = typedRoute.route.fwdAzimuth;
            if (!backAzimuth) backAzimuth = typedRoute.route.backAzimuth;
          }
        });
      } catch (error: any) {
        throw new Error('Invalid route');
      }

      const ad = AdEntity.create({
        id: command.id,
        driver: command.driver,
        passenger: command.passenger,
        frequency: command.frequency,
        fromDate: command.fromDate,
        toDate: command.toDate,
        schedule: command.schedule,
        seatsProposed: command.seatsProposed,
        seatsRequested: command.seatsRequested,
        strict: command.strict,
        waypoints: command.waypoints,
        points: points as PointValueObject[],
        driverDistance,
        driverDuration,
        passengerDistance,
        passengerDuration,
        fwdAzimuth: fwdAzimuth as number,
        backAzimuth: backAzimuth as number,
      });

      try {
        await this.repository.insertExtra(ad, 'ad');
        return ad.id;
      } catch (error: any) {
        if (error instanceof ConflictException) {
          throw new AdAlreadyExistsException(error);
        }
        throw error;
      }
    } catch (error: any) {
      const matcherAdCreationFailedIntegrationEvent =
        new MatcherAdCreationFailedIntegrationEvent({
          id: command.id,
          metadata: {
            correlationId: command.id,
            timestamp: Date.now(),
          },
          cause: error.message,
        });
      this.messagePublisher.publish(
        MATCHER_AD_CREATION_FAILED_ROUTING_KEY,
        JSON.stringify(matcherAdCreationFailedIntegrationEvent),
      );
      throw error;
    }
  }
}
